#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
#
# Postfix mail log files parser that inserts information about
# mail traffic into the SQL database
#
# Copyright (c) 2006 JAS
#
# Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
#
# $Id: mailsql.py 6 2007-01-27 04:38:24Z vokac $
#
"""This program parse postfix log files and inserts information about
mail traffic into the SQL database"""
import os, sys, signal, stat, time, getopt, re
import logging, logging.handlers
import MySQLdb

__version__ = "$Revision: 6 $"

# global variables
quitObjects = []

class LogCat:

    # wait in case reach end of the file
    nextInterval = 1
    # wait in case of some exception
    errorInterval = 10
    # check that file was rotated
    checkInterval = 60

    def __init__(self, fileName):
        if fileName == None:
            raise Exception("filename must be defined")
        self._fileName = fileName
        self._file = None
        self._inode = None
        self._stop = False

    def _sleep(self, timeout):
        last = time.time()
        while timeout > 0 and not self._stop:
            time.sleep(1)
            now = time.time()
            delta = now - last
            if delta < 0 or delta > 1:
                delta = 1
            timeout -= delta
            last = now

    def stop(self):
        self._stop = True

    def cat(self):
        """Call processLine function for each new line in log file."""
        while not self._stop:
            try:
                if not os.path.isfile(self._fileName):
                    raise Exception("file %s doesn't exist" % self._fileName)
                if self._file == None:
                    # open file
                    try:
                        logging.getLogger().debug("opening file %s" % self._fileName)
                        self._file = open(self._fileName)
                        self._inode = os.fstat(self._file.fileno())[stat.ST_INO]
                    except IOError, e:
                        raise Exception("can't open file %s (%s)" % (self._fileName, e))
                else:
                    # check if file was rotated
                    try:
                        if self._inode != os.stat(self._fileName)[stat.ST_INO]:
                            logging.getLogger().info("file %s changed" % self._fileName)
                            self._file = None
                            self._inode = None
                            continue
                    except OSError, e:
                        raise Exception("can't stat file %s (%s)" % (self._fileName, e))
                # parse log file
                last = time.time()
                checkTime = LogCat.checkInterval
                while not self._stop and checkTime > 0:
                    now = time.time()
                    delta = now - last
                    last = now
                    if delta < 0: delta = 0
                    checkTime -= delta

                    where = self._file.tell()
                    line = self._file.readline()
                    if not line:
                        self._file.seek(where)
                        self._sleep(LogCat.nextInterval)
                    else:
                        self.processLine(line)

            except Exception, e:
                logging.getLogger().error(e)
                import traceback
                exc_info_type, exc_info_value, exc_info_traceback = sys.exc_info()
                logging.getLogger().error("".join(traceback.format_exception(exc_info_type, exc_info_value, exc_info_traceback)))
                self._file = None
                self._inode = None
                self._sleep(LogCat.errorInterval)

    def processLine(self, line):
        raise Exception("this method has to be implemented in children")



class MailLogCat(LogCat):

    monthMap = { 'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4,
                 'may': 5, 'jun': 6, 'jul': 7, 'aug': 8,
                 'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12 }

    def __init__(self, fileName, **keywords):
        LogCat.__init__(self, fileName)
        self.db_host = keywords.get('db_host', 'localhost')
        self.db_port = keywords.get('db_port', '3306')
        self.db_db = keywords.get('db_db', 'maillog')
        self.db_user = keywords.get('db_user', 'maillog')
        self.db_pass = keywords.get('db_pass', 'secret')
        self.conn = None
        self.cache = {}
        self.cacheEx = {}
        self.cacheMax = 10000
        self.cacheExpire = 60*60*24
        self.reBegin = re.compile("^(\S+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\S+)\s+")
        self.reAppl = re.compile("^([^[]+)\[([^]]+)\]: ")
        while True:
            cursor = None
            try:
                if self.conn == None:
                    self.conn = MySQLdb.connect(host = self.db_host,
                                                db = self.db_db,
                                                user = self.db_user,
                                                passwd = self.db_pass)
                cursor = self.conn.cursor()
                query = "SELECT IFNULL(UNIX_TIMESTAMP(MAX(time_iso)), 0) AS skipTime FROM log_mail"
                logging.getLogger().debug("sql: %s" % query)
                cursor.execute(query)
                res = cursor.fetchone()
                if res == None:
                    self.skip = None
                else:
                    logging.getLogger().info("skipping to time: %s (%s), current time: %s (%s)" % (res[0], time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(res[0])), time.time(), time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))))
                    self.skip = res[0]
                cursor.close()
                break
            except MySQLdb.DatabaseError, err:
                logging.getLogger().error("database error: %s" % err)
                if self.conn != None:
                    try:
                        if cursor == None:
                            cursor = self.conn.cursor()
                        cursor.execute('SELECT 1')
                        cursor.close()
                    except Exception, e:
                        try:
                            self.conn.close()
                        finally:
                            self.conn = None
                self._sleep(LogCat.errorInterval)

    def processLine(self, line):
        begin = self.reBegin.search(line)
        if begin == None:
            if logging.getLogger().getEffectiveLevel() <= logging.WARN: logging.getLogger().warn("unknown line format: %s" % line.rstrip())
            return
        year = time.localtime()[0] # XXXX
        month = MailLogCat.monthMap.get(begin.group(1).lower(), 0)
        day = int(begin.group(2))
        hour = int(begin.group(3))
        minut = int(begin.group(4))
        sec = int(begin.group(5))
        host = begin.group(6)
        time_iso = int(time.mktime((year, month, day, hour, minut, sec, 0, 0, 0)))
        if time.time() + 86400 < time_iso:
            # time in future?!
            year -= 1
            time_iso = int(time.mktime((year, month, day, hour, minut, sec, 0, 0, 0)))
        if time.localtime(time_iso)[-1] == 1:
            # Daylight Savings Time flag is set
            time_iso += time.timezone
        if self.skip != None and time_iso <= self.skip:
            # if logging.getLogger().getEffectiveLevel() <= logging.DEBUG: logging.getLogger().debug("skipping line: %s" % line.rstrip())
            return
        appl = self.reAppl.search(line[begin.end():])
        if appl == None:
            if logging.getLogger().getEffectiveLevel() <= logging.DEBUG: logging.getLogger().debug("unknown line appl: %s" % line.rstrip())
            return
        app = appl.group(1)
        pid = appl.group(2)
        if app.find('/') != -1:
            app, proc = app.split('/', 1)
        else:
            proc = ''
        if app != 'postfix':
            # logging.getLogger().debug("we are not interested in line: %s" % line)
            return

        text = line[begin.end()+appl.end():].rstrip()

        if proc == 'smtpd':
            a = re.compile("^NOQUEUE: reject: .*?\[(.*?)\]: (\d+).*?: (.*); from=<(.*?)> to=<(.*?)> proto=([^ ]*) helo=<(.*?)>")
            m = a.search(text)
            if m != None:
                mail_id = self.insert(
                    'log_mail',
                    { 'time_iso': time_iso, 'queue_id': 'NOT_QUEUED',
                      'host': host, 'client': m.group(1),
                      'helo': m.group(7), 'sender': m.group(4),
                      'proto': m.group(6) }, time_iso)
                if mail_id == None:
                    return
                self.insert(
                    'log_recip',
                    { 'mail_id': mail_id, 'email': m.group(5),
                      'code': m.group(2), 'status': 'reject',
                      'info': m.group(3)[:254] }, time_iso)
                return
            a = re.compile("^NOQUEUE: filter: .*?\[(.*?)\]: (.*); from=<(.*?)> to=<(.*?)> proto=([^ ]*) helo=<(.*?)>")
            m = a.search(text)
            if m != None:
                mail_id = self.insert(
                    'log_mail',
                    { 'time_iso': time_iso, 'queue_id': 'FILTER',
                      'host': host, 'client': m.group(1),
                      'helo': m.group(6), 'sender': m.group(3),
                      'proto': m.group(5) }, time_iso)
                if mail_id == None:
                    return
                self.insert(
                    'log_recip',
                    { 'mail_id': mail_id, 'email': m.group(4),
                      'code': 0, 'status': 'filter',
                      'info': m.group(2)[:254] }, time_iso)
            a = re.compile("^([0-9A-F]+): client=.*?\[(.*?)\]")
            m = a.search(text)
            if m != None:
                res = self.select('log_mail', [ 'id' ], { 'host': host, 'queue_id': 'FILTER', 'client': m.group(2) }, time_iso)
                if res != None and len(res) > 0:
                    mail_id = self.update( 'log_mail',
                        { 'time_iso': time_iso, 'queue_id': m.group(1) },
                        { 'host': host, 'queue_id': 'FILTER', 'client': m.group(2) },
                        time_iso)
                else:
                    mail_id = self.insert(
                        'log_mail',
                        { 'time_iso': time_iso, 'queue_id': m.group(1),
                          'host': host, 'client': m.group(2) }, time_iso)
                return
            return

        if proc == 'smtp' or proc == 'local':
            a = re.compile("^([0-9A-F]+): to=<(.*?)>(.*), relay=(\S+)(.*), delay=(\d+)(.*), status=(\S+) (.*)")
            m = a.search(text)
            if m != None:
                queue_id = m.group(1)
                mail_id = self.getMailId((host, queue_id), time_iso)
                if mail_id == None:
                    logging.getLogger().info("can't find mail_id for %s/%s/%s" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time_iso)), host, queue_id))
                    return
                recip = m.group(2)
                recip_orig = ''
                if m.group(3)[:len(', orig_to=<')] == ', orig_to=<':
                    recip_orig = m.group(3)[len(', orig_to=<'):-1]
                relay = m.group(4)
                status = m.group(8)
                info = m.group(9)
                code = 0
                if re.search("^\((\d+)", info) != None:
                    code = re.search("^\((\d+)", info).group(1)
                self.insert(
                    'log_recip',
                    { 'mail_id': mail_id, 'email': recip,
                      'email_orig': recip_orig, 'relay': relay,
                      'code': code, 'status': status,
                      'info': info[:254] }, time_iso)
                return
            return

        if proc == 'qmgr':
            a = re.compile("^([0-9A-F]+): from=<(.*?)>, ")
            m = a.search(text)
            if m != None:
                queue_id = m.group(1)
                sender = m.group(2)
                m1 = re.search("size=(\d+), nrcpt=(\d+)", text[m.end():])
                if m1 != None:
                    size = m1.group(1)
                    nrcpt = m1.group(2)
                    self.update('log_mail',
                                { 'time_iso': time_iso, 'sender': sender,
                                  'size': size, 'nrcpt': nrcpt },
                                { 'queue_id': queue_id, 'host': host },
                                time_iso)
                    return
                m1 = re.search("status=(\S+?), (.*)", text[m.end():])
                if m1 != None:
                    status = m1.group(1)
                    info = m1.group(2)
                    mail_id = self.getMailId((host, queue_id), time_iso)
                    if mail_id == None:
                        logging.getLogger().info("can't find mail_id for %s/%s/%s" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time_iso)), host, queue_id))
                        return
                    if status.lower() == 'expired':
                        self.update('log_recip',
                                    { 'status': 'expired' },
                                    { 'mail_id': mail_id },
                                    time_iso)
                    else:
                        self.update('log_recip',
                                    { 'status': status, 'info': info[:254] },
                                    { 'mail_id': mail_id },
                                    time_iso)
                    return
            return

        if proc == 'pickup':
            a = re.compile("^([0-9A-F]+): uid=(\d+) from=<(.*?)>")
            m = a.search(text)
            if m != None:
                mail_id = self.insert(
                    'log_mail',
                    { 'time_iso': time_iso, 'queue_id': m.group(1),
                      'host': host, 'client': 'unix',
                      'sender': m.group(3), 'proto': "uid=%s" % m.group(2) }, time_iso)
                return
            return

        if proc == 'cleanup':
            a = re.compile("^([0-9A-F]+): milter-(?:reject|discard): .*?\[(.*?)\]: (.*); from=<(.*?)> to=<(.*?)> proto=([^ ]*) helo=<(.*?)>")
            m = a.search(text)
            if m != None:
                mail_id = self.insert(
                    'log_mail',
                    { 'time_iso': time_iso, 'queue_id': m.group(1),
                      'host': host, 'client': m.group(2),
                      'helo': m.group(7), 'sender': m.group(4),
                      'proto': m.group(6) }, time_iso)
                if mail_id == None:
                    return
                self.insert(
                    'log_recip',
                    { 'mail_id': mail_id, 'email': m.group(5),
                      'code': 450, 'status': 'reject',
                      'info': m.group(3)[:254] }, time_iso)
                return
            a = re.compile("^([0-9A-F]+): message-id=<(.*?)>")
            m = a.search(text)
            if m != None:
                res = self.select('log_mail', [ 'id' ], { 'host': host, 'queue_id': m.group(1) }, time_iso)
                if res == None or len(res) == 0:
                    mail_id = self.insert(
                        'log_mail',
                        { 'time_iso': time_iso, 'queue_id': m.group(1),
                          'host': host, 'client': 'local' }, time_iso)
                return
            return

    def getMailId(self, key, time_iso):
        mail_id = self.cache.get(key)
        expire = self.cacheEx.get(key)
        if mail_id == None or expire == None or expire + 86400 < time_iso:
            # try to find in DB - should happend only after restart
            host = key[0]
            queue_id = key[1]
            res = self.select('log_mail', [ 'id' ], { 'host': host, 'queue_id': queue_id }, time_iso)
            if res != None and len(res) > 0:
                self.setMailId(key, res[0][0], time_iso)
                return res[0][0]
        return mail_id


    def setMailId(self, key, value, time_iso):
        if len(self.cache) > self.cacheMax:
            # cache cleanup
            expire = self.cacheEx.values()
            expire.sort()
            expireTr = expire[int(self.cacheMax*1/4)]
            if expireTr < time.time() - self.cacheExpire:
                expireTr = time.time() - self.cacheExpire
            toDel = []
            for k, v in self.cacheEx.items():
                if v <= expireTr or v >= time.time() + self.cacheExpire:
                    toDel.append(k)
            for k in toDel:
                del(self.cache[k])
                del(self.cacheEx[k])

        self.cache[key] = value
        self.cacheEx[key] = time.time()


    def insert(self, table, data, time_iso):
        keys = "`,`".join(data.keys())
        vals = []
        for k in data.keys():
            if k.lower() == 'time_iso':
                vals.append("FROM_UNIXTIME(%s)")
            else:
                vals.append("%s")
        query = "INSERT INTO `%s` (`%s`) VALUES (%s)" % (table, keys, ",".join(vals))
        ret = self.query(query, tuple(data.values()))
        if ret != None:
            if data.has_key('host') and data.has_key('queue_id'):
                self.setMailId((data['host'], data['queue_id']), ret[0][0], time_iso)
            return ret[0][0]
        else:
            return None


    def update(self, table, data, cond, time_iso, interval = 5*86400):
        vals = []
        for k in data.keys():
            if k.lower() == 'time_iso':
                vals.append("%s = FROM_UNIXTIME(%%s)" % k)
            else:
                vals.append("%s = %%s" % k)
        conds = []
        for k in cond.keys():
            if k.lower() == 'time_iso':
                conds.append("%s = FROM_UNIXTIME(%%s)" % k)
            else:
                conds.append("%s = %%s" % k)
        # FIXME: here should be something like time_iso - db_time_iso
        if table.lower() == 'log_mail':
            conds.append("%s - UNIX_TIMESTAMP(`time_iso`) BETWEEN 0 AND %s" % (time_iso, interval))
        values = ",".join(vals)
        where = " AND ".join(conds)
        query = "UPDATE `%s` SET %s WHERE %s" % (table, values, where)
        ret = None
        if data.has_key('queue_id') and cond.has_key('host'):
            ret = self.select(table, [ 'id' ], cond, time_iso)
            if ret != None and len(ret) > 0:
                ret = ret[0][0]
                self.setMailId((cond['host'], data['queue_id']), ret, time_iso)
        self.query(query, tuple(data.values() + cond.values()))
        return ret


    def select(self, table, cols, cond, time_iso, interval = 5*86400):
        columns = "`,`".join(cols)
        conds = []
        for k in cond.keys():
            if k.lower() == 'time_iso':
                conds.append("%s = FROM_UNIXTIME(%%s)" % k)
            else:
                conds.append("%s = %%s" % k)
        if table.lower() == 'log_mail':
            conds.append("%s - UNIX_TIMESTAMP(`time_iso`) BETWEEN 0 AND %s" % (time_iso, interval))
        where = " AND ".join(conds)
        query = "SELECT `%s` FROM `%s` WHERE %s" % (columns, table, where)
        return self.query(query, tuple(cond.values()))


    def query(self, query, data = ()):
        retry = 5
        while retry > 0:
            sleep = 0
            retry -= 1
            cursor = None
            try:
                if self.conn == None:
                    self.conn = MySQLdb.connect(host = self.db_host,
                                                db = self.db_db,
                                                user = self.db_user,
                                                passwd = self.db_pass)
                cursor = self.conn.cursor()
                if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
                    logging.getLogger().debug("sql: %s" % query)
                    logging.getLogger().debug("data: %s" % str(data))
                cursor.execute(query, data)
                if query[:len('select')].lower() != 'select':
                    cursor.execute('COMMIT')
                    # cursor.execute('SET AUTOCOMMIT = 1'
                if query[:len('insert')].lower() == 'insert':
                    cursor.execute('SELECT LAST_INSERT_ID()')
                res = cursor.fetchall()
                cursor.close()
                return res
            except MySQLdb.DatabaseError, err:
                logging.getLogger().error("database error: %s" % err)
                if self.conn != None:
                    try:
                        if cursor == None:
                            cursor = self.conn.cursor()
                        cursor.execute('SELECT 1')
                        cursor.close()
                    except Exception, e:
                        try:
                            self.conn.close()
                        finally:
                            self.conn = None
                self._sleep(sleep)
                sleep += 1
        return None



def daemonize():
    # do the UNIX double-fork magic, see Stevens' "Advanced
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError, e:
        logging.getLogger().error("fork #1 failed: %d (%s)" % (e.errno, e.strerror))
        sys.exit(1)

    # decouple from parent environment
    os.chdir("/")   #don't prevent unmounting....
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent, print eventual PID before
            logging.getLogger().info("daemon PID %d" % pid)
            sys.exit(0)
    except OSError, e:
        logging.getLogger().error("fork #2 failed: %d (%s)" % (e.errno, e.strerror))
        sys.exit(1)



def quitHandler(signum, frame):
    global quitObjects
    logging.getLogger().info("received signal %s (%s), terminating..." % (signum, frame))
    for obj in quitObjects:
        obj.stop()
#    sys.exit(0)



def usage():
    print __doc__
    print "usage: %s [-h] [-v] [-q] [-d] -f maillog" % sys.argv[0]
    print "  -h, --help         display this help and exit"
    print "  -q, --quiet        be quiet about what you do"
    print "  -v, --verbose num  be verbose about what you do (num - log level)"
    print "  -V, --version      output version information and exit"
    print "  -d, --daemon       start in the background"
    print "  --logfile=FILE     monitor defined logfile instead of /var/log/maillog"
    print "  --daemon-pid=FILE  write PID to FILE instead of /var/run/mailsql.pid"
    print "  --daemon-log=FILE  write verbose-log to FILE instead of /var/log/mailsql.log"
    print "  --db-host=hostname SQL database hostname (localhost)"
    print "  --db-port=port     SQL database port (3306)"
    print "  --db-name=name     SQL database name (maillog)"
    print "  --db-user=username SQL database user (maillog)"
    print "  --db-pass=password SQL database password (secret)"



if __name__ == "__main__":
    streamHandler = logging.StreamHandler(sys.stdout)
    streamHandler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s](%(module)s:%(lineno)d) %(message)s", "%d %b %H:%M:%S"))
    logging.getLogger().addHandler(streamHandler)
    logging.getLogger().setLevel(logging.WARN)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hqvVdl:",
                ["help", "quiet", "verbose", "version"
                 "daemon", "logfile=", "log-level=", "daemon-pid=",
                 "daemon-log=", "db-host=", "db-port=",
                 "db-name=", "db-user=", "db-pass=" ])
    except getopt.GetoptError, e:
        logging.getLogger().warn("command line error: %s" % e)
        usage()
        sys.exit(1)

    logfile = "/var/log/maillog"
    daemon = False
    daemonLogFile = "/var/log/mailsql.log"
    db_host = "localhost"
    db_port = "3306"
    db_db = "maillog"
    db_user = "maillog"
    db_pass = "secret"
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-q", "--quiet"):
            logging.getLogger().setLevel(logging.ERROR)
        if o in ("-v", "--verbose"):
            logging.getLogger().setLevel(logging.DEBUG)
        if o == "--log-level":
            logging.getLogger().setLevel(int(a))
        if o in ("-V", "--version"):
            print "%s version: %s" % (sys.argv[0], __version__)
        if o in ("-l", "--logfile"):
            logfile = a
        if o == "--daemon-log":
            daemonLogFile = a
        if o in ("-d", "--daemon"):
            daemon = True
        if o == "--db-host":
            db_host = a
        if o == "--db-port":
            db_port = a
        if o == "--db-name":
            db_db = a
        if o == "--db-user":
            db_user = a
        if o == "--db-pass":
            db_pass = a

    if logfile == None:
        usage()
        sys.exit(1)

    if daemon:
        if daemonLogFile != None and daemonLogFile != '':
            # FIXME: don't overide
#            fileHandler = logging.handlers.TimedRotatingFileHandler(daemonLogFile, 'midnight', 1, 4)
            fileHandler = logging.handlers.RotatingFileHandler(daemonLogFile, maxBytes=10*1024*1024, backupCount=4)
            fileHandler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s](%(module)s:%(lineno)d) %(message)s", "%d %b %H:%M:%S"))
            logging.getLogger().addHandler(fileHandler)
        logging.getLogger().removeHandler(streamHandler)
        daemonize()
        signal.signal(signal.SIGTERM, quitHandler)
#        if daemonPidFile != None and daemonPidFile != '':
#            pidFile = open(daemonPidFile, "w")
#            pidFile.write("%s" % os.getpid())
#            pidFile.close()
    else:
        signal.signal(signal.SIGINT, quitHandler)

    logging.getLogger().info("command: %s" % " ".join(sys.argv))
    logging.getLogger().info("version: %s" % __version__)

    log = MailLogCat(logfile, db_host=db_host, db_port=db_port,
                     db_db=db_db, db_user=db_user, db_pass=db_pass)
    quitObjects.append(log)
    log.cat()
